import { ReactiveMap } from "@solid-primitives/map";
import { type mastodon } from "masto";
import {
  Accessor,
  batch,
  catchError,
  createEffect,
  createResource,
  untrack,
  type Resource,
  type ResourceFetcherInfo,
} from "solid-js";
import { createStore } from "solid-js/store";

type Timeline<T extends mastodon.DefaultPaginationParams> = {
  list(params?: T): mastodon.Paginator<mastodon.v1.Status[], unknown>;
};

type TimelineParamsOf<T> = T extends Timeline<infer P> ? P : never;

export function createTimelineControlsForArray(
  status: () => mastodon.v1.Status[] | undefined,
): TimelineControls {
  const lookup = new ReactiveMap<string, TreeNode<mastodon.v1.Status>>();

  const [threads, setThreads] = createStore([] as mastodon.v1.Status["id"][]);

  createEffect(() => {
    const nls = catchError(status, (e) => {
      console.error(e);
    });
    if (!nls) return;

    setThreads([]);
    lookup.clear();

    const existence = [] as boolean[];

    for (const [idx, status] of nls.entries()) {
      existence[idx] = !!untrack(() => lookup.get(status.id));
      lookup.set(status.id, {
        value: status,
      });
    }

    for (const status of nls) {
      const node = untrack(() => lookup.get(status.id))!;
      if (status.inReplyToId) {
        const parent = lookup.get(status.inReplyToId);
        if (parent) {
          const children = parent.children ?? [];
          if (!children.find((x) => x.value.id == status.id)) {
            children.push(node);
          }
          parent.children = children;
          node.parent = parent;
        }
      }
    }

    const newThreads = nls
      .filter((x, i) => !existence[i])
      .map((x) => x.id)
      .filter((id) => (lookup.get(id)!.children?.length ?? 0) === 0);

    setThreads(newThreads);
  });

  return {
    list: threads,
    get(id: string) {
      return lookup.get(id);
    },
    getPath(id: string) {
      const node = lookup.get(id);
      if (!node) return;
      return collectPath(node);
    },
    set(id: string, value: mastodon.v1.Status) {
      const node = untrack(() => lookup.get(id));
      if (!node) return;
      node.value = value;
      lookup.set(id, node);
    },
  };
}

export function createTimelineSnapshot<
  T extends Timeline<mastodon.DefaultPaginationParams>,
>(
  timeline: Accessor<T>,
  params: Accessor<TimelineParamsOf<T>>,
): TimelineResource<mastodon.v1.Status[] | undefined> {
  const [shot, { refetch }] = createResource(
    () => [timeline(), params()] as const,
    async ([tl, limit]) => {
      const ls = await tl.list(limit).next();
      return ls.value;
    },
  );

  const controls = createTimelineControlsForArray(shot);

  return [
    controls,
    shot,
    {
      refetch,
    },
  ] as const;
}

export type TimelineFetchDirection = mastodon.Direction;

export type TimelineChunk<T extends mastodon.DefaultPaginationParams> = {
  tl: Timeline<T>;
  rebuilt: boolean;
  chunk: readonly mastodon.v1.Status[];
  done?: boolean;
  direction: TimelineFetchDirection;
  params: T;
};

type TreeNode<T> = {
  parent?: TreeNode<T>;
  value: T;
  children?: TreeNode<T>[];
};

/** Collect the path of a node for the root.
 * The first element is the node itself, the last element is the root.
 */
function collectPath<T>(node: TreeNode<T>) {
  const path = [node] as TreeNode<T>[];
  let current = node;
  while (current.parent) {
    path.push(current.parent);
    current = current.parent;
  }
  return path;
}

function createTimelineChunk<
  T extends Timeline<mastodon.DefaultPaginationParams>,
>(timeline: Accessor<T>, params: Accessor<TimelineParamsOf<T>>) {
  let vpMaxId: string | undefined, vpMinId: string | undefined;

  const fetchExtendingPage = async (
    tl: T,
    direction: TimelineFetchDirection,
    params: TimelineParamsOf<T>,
  ) => {
    switch (direction) {
      case "next": {
        const page = await tl
          .list({ ...params, sinceId: vpMaxId })
          .setDirection(direction)
          .next();
        if ((page.value?.length ?? 0) > 0) {
          vpMaxId = page.value![0].id;
        }
        return page;
      }

      case "prev": {
        const page = await tl
          .list({ ...params, maxId: vpMinId })
          .setDirection(direction)
          .next();
        if ((page.value?.length ?? 0) > 0) {
          vpMinId = page.value![page.value!.length - 1].id;
        }
        return page;
      }
    }
  };

  return createResource(
    () => [timeline(), params()] as const,
    async (
      [tl, params],
      info: ResourceFetcherInfo<
        Readonly<TimelineChunk<TimelineParamsOf<T>>>,
        TimelineFetchDirection
      >,
    ) => {
      const direction =
        typeof info.refetching === "boolean" ? "prev" : info.refetching;
      const rebuildTimeline = tl !== info.value?.tl;
      if (rebuildTimeline) {
        vpMaxId = undefined;
        vpMinId = undefined;
      }
      const posts = await fetchExtendingPage(tl, direction, params);
      return {
        tl,
        rebuilt: rebuildTimeline,
        chunk: posts.value ?? [],
        done: posts.done,
        direction,
        params,
      };
    },
  );
}

export type TimelineControls = {
  /**
   * The threads.
   *
   * The identifiers here is the most-bottom toot id in the thread.
   *
   * @see You can use {@link TimelineControls.get} and {@link TimelineControls.getPath} to resolve them if
   * the context is needed.
   */
  list: readonly mastodon.v1.Status["id"][];
  /**
   * Get the single node.
   */
  get(id: string): TreeNode<mastodon.v1.Status> | undefined;
  /**
   * Collect the path from the node to the most-top node.
   */
  getPath(id: string): TreeNode<mastodon.v1.Status>[] | undefined;
  /**
   * Set the node value.
   */
  set(id: string, value: mastodon.v1.Status): void;
};

export type TimelineResource<R> = [
  TimelineControls,
  Resource<R>,
  { refetch(info?: TimelineFetchDirection): void },
];

/**
 * Create auto managed timeline controls.
 *
 * The error from the resource is not thrown in the
 * {@link TimelineControls["list"]} and {@link TimelineControls}.get*.
 * Use the second value from {@link TimelineResource} to catch the error.
 */
export function createTimeline<
  T extends Timeline<mastodon.DefaultPaginationParams>,
>(
  timeline: Accessor<T>,
  params: Accessor<TimelineParamsOf<T>>,
): TimelineResource<TimelineChunk<TimelineParamsOf<T>> | undefined> {
  const lookup = new ReactiveMap<string, TreeNode<mastodon.v1.Status>>();
  const [threads, setThreads] = createStore([] as mastodon.v1.Status["id"][]);

  const [chunk, { refetch }] = createTimelineChunk(timeline, params);

  createEffect(() => {
    const chk = catchError(chunk, (e) => console.error(e));
    if (!chk) {
      return;
    }

    if (chk.rebuilt) {
      lookup.clear();
      setThreads([]);
    }

    const existence = [] as boolean[];

    for (const [idx, status] of chk.chunk.entries()) {
      existence[idx] = !!untrack(() => lookup.get(status.id));
      lookup.set(status.id, {
        value: status,
      });
    }

    for (const status of chk.chunk) {
      const node = untrack(() => lookup.get(status.id))!;
      if (status.inReplyToId) {
        const parent = lookup.get(status.inReplyToId);
        if (parent) {
          const children = parent.children ?? [];
          if (!children.find((x) => x.value.id == status.id)) {
            children.push(node);
          }
          parent.children = children;
          node.parent = parent;
        }
      }
    }

    const nThreadIds = chk.chunk
      .filter((x, i) => !existence[i])
      .map((x) => x.id);

    batch(() => {
      if (chk.direction === "prev") {
        setThreads((threads) => [...threads, ...nThreadIds]);
      } else if (chk.direction === "next") {
        setThreads((threads) => [...nThreadIds, ...threads]);
      }

      setThreads((threads) =>
        threads.filter((id) => (lookup.get(id)!.children?.length ?? 0) === 0),
      );
    });
  });

  return [
    {
      list: threads,
      get(id: string) {
        return lookup.get(id);
      },
      getPath(id: string) {
        const node = lookup.get(id);
        if (!node) return;
        return collectPath(node);
      },
      set(id: string, value: mastodon.v1.Status) {
        const node = untrack(() => lookup.get(id));
        if (!node) return;
        node.value = value;
        lookup.set(id, node);
      },
    },
    chunk,
    { refetch },
  ] as const;
}
